'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var stripAnsi = _interopDefault(require('strip-ansi'));
var chalk = _interopDefault(require('chalk'));
var path = _interopDefault(require('path'));
var EventEmitter = _interopDefault(require('events'));
var minimost = _interopDefault(require('minimost'));
var readPkg = _interopDefault(require('read-pkg-up'));

var textTable = function (rows_, opts) {
    if (!opts) opts = {};
    var hsep = opts.hsep === undefined ? '  ' : opts.hsep;
    var align = opts.align || [];
    var stringLength = opts.stringLength
        || function (s) { return String(s).length; }
    ;
    
    var dotsizes = reduce(rows_, function (acc, row) {
        forEach(row, function (c, ix) {
            var n = dotindex(c);
            if (!acc[ix] || n > acc[ix]) acc[ix] = n;
        });
        return acc;
    }, []);
    
    var rows = map(rows_, function (row) {
        return map(row, function (c_, ix) {
            var c = String(c_);
            if (align[ix] === '.') {
                var index = dotindex(c);
                var size = dotsizes[ix] + (/\./.test(c) ? 1 : 2)
                    - (stringLength(c) - index)
                ;
                return c + Array(size).join(' ');
            }
            else return c;
        });
    });
    
    var sizes = reduce(rows, function (acc, row) {
        forEach(row, function (c, ix) {
            var n = stringLength(c);
            if (!acc[ix] || n > acc[ix]) acc[ix] = n;
        });
        return acc;
    }, []);
    
    return map(rows, function (row) {
        return map(row, function (c, ix) {
            var n = (sizes[ix] - stringLength(c)) || 0;
            var s = Array(Math.max(n + 1, 1)).join(' ');
            if (align[ix] === 'r' || align[ix] === '.') {
                return s + c;
            }
            if (align[ix] === 'c') {
                return Array(Math.ceil(n / 2 + 1)).join(' ')
                    + c + Array(Math.floor(n / 2 + 1)).join(' ')
                ;
            }
            
            return c + s;
        }).join(hsep).replace(/\s+$/, '');
    }).join('\n');
};

function dotindex (c) {
    var m = /\.[^.]*$/.exec(c);
    return m ? m.index + 1 : c.length;
}

function reduce (xs, f, init) {
    if (xs.reduce) return xs.reduce(f, init);
    var i = 0;
    var acc = arguments.length >= 3 ? init : xs[i++];
    for (; i < xs.length; i++) {
        f(acc, xs[i], i);
    }
    return acc;
}

function forEach (xs, f) {
    if (xs.forEach) return xs.forEach(f);
    for (var i = 0; i < xs.length; i++) {
        f.call(xs, xs[i], i);
    }
}

function map (xs, f) {
    if (xs.map) return xs.map(f);
    var res = [];
    for (var i = 0; i < xs.length; i++) {
        res.push(f.call(xs, xs[i], i));
    }
    return res;
}

/* eslint-disable yoda */
var isFullwidthCodePoint = x => {
	if (Number.isNaN(x)) {
		return false;
	}

	// code points are derived from:
	// http://www.unix.org/Public/UNIDATA/EastAsianWidth.txt
	if (
		x >= 0x1100 && (
			x <= 0x115f ||  // Hangul Jamo
			x === 0x2329 || // LEFT-POINTING ANGLE BRACKET
			x === 0x232a || // RIGHT-POINTING ANGLE BRACKET
			// CJK Radicals Supplement .. Enclosed CJK Letters and Months
			(0x2e80 <= x && x <= 0x3247 && x !== 0x303f) ||
			// Enclosed CJK Letters and Months .. CJK Unified Ideographs Extension A
			(0x3250 <= x && x <= 0x4dbf) ||
			// CJK Unified Ideographs .. Yi Radicals
			(0x4e00 <= x && x <= 0xa4c6) ||
			// Hangul Jamo Extended-A
			(0xa960 <= x && x <= 0xa97c) ||
			// Hangul Syllables
			(0xac00 <= x && x <= 0xd7a3) ||
			// CJK Compatibility Ideographs
			(0xf900 <= x && x <= 0xfaff) ||
			// Vertical Forms
			(0xfe10 <= x && x <= 0xfe19) ||
			// CJK Compatibility Forms .. Small Form Variants
			(0xfe30 <= x && x <= 0xfe6b) ||
			// Halfwidth and Fullwidth Forms
			(0xff01 <= x && x <= 0xff60) ||
			(0xffe0 <= x && x <= 0xffe6) ||
			// Kana Supplement
			(0x1b000 <= x && x <= 0x1b001) ||
			// Enclosed Ideographic Supplement
			(0x1f200 <= x && x <= 0x1f251) ||
			// CJK Unified Ideographs Extension B .. Tertiary Ideographic Plane
			(0x20000 <= x && x <= 0x3fffd)
		)
	) {
		return true;
	}

	return false;
};

var stringWidth = str => {
	if (typeof str !== 'string' || str.length === 0) {
		return 0;
	}

	str = stripAnsi(str);

	let width = 0;

	for (let i = 0; i < str.length; i++) {
		const code = str.codePointAt(i);

		// Ignore control characters
		if (code <= 0x1F || (code >= 0x7F && code <= 0x9F)) {
			continue;
		}

		// Ignore combining characters
		if (code >= 0x300 && code <= 0x36F) {
			continue;
		}

		// Surrogates
		if (code > 0xFFFF) {
			i++;
		}

		width += isFullwidthCodePoint(code) ? 2 : 1;
	}

	return width;
};

class CacError extends Error {
    constructor(message) {
        super(message);
        this.name = 'CacError';
    }
}

function orderNames(names) {
    return names.sort((a, b) => {
        return a.length > b.length ? 1 : -1;
    });
}
function textTable$1(data) {
    return textTable(data, {
        stringLength: stringWidth
    });
}
function prefixOption(option) {
    return option.length === 1 ? `-${option}` : `--${option}`;
}
function isExplictCommand(name) {
    return name && !name.startsWith('-');
}
function invariant(exp, message) {
    if (!exp) {
        throw new CacError(message);
    }
}

class Options {
    constructor() {
        this.options = [];
    }
    add(name, opt) {
        let names = [name];
        if (typeof opt === 'string') {
            opt = { desc: opt };
        }
        else if (typeof opt === 'object') {
            names = names.concat(opt.alias || []);
        }
        const option = Object.assign({}, opt, { name, names: orderNames(names) });
        this.options.push(option);
        return this;
    }
    getDefaultsMapping() {
        return this.options
            .filter(option => {
            return typeof option.default !== 'undefined';
        })
            .reduce((res, next) => {
            if (next.default !== undefined) {
                res[next.name] = next.default;
            }
            return res;
        }, {});
    }
    getOptionsByType(type) {
        return this.options.filter(option => type === option.type);
    }
    getOptionNamesByType(type) {
        return this.getOptionsByType(type).map(option => option.name);
    }
    getAliasMap() {
        return this.options.reduce((res, next) => {
            if (next.alias !== undefined) {
                res[next.name] = next.alias;
            }
            return res;
        }, {});
    }
    isEmpty() {
        return this.options.length === 0;
    }
    toString() {
        return textTable$1(this.options.map(option => {
            const extra = [];
            if (typeof option.default !== 'undefined') {
                extra.push(`Default: ${JSON.stringify(option.default)}`);
            }
            if (typeof option.type === 'string') {
                extra.push(`Type: ${option.type}`);
            }
            return [
                option.names.map(v => chalk.yellow(prefixOption(v))).join(', '),
                chalk.dim(option.desc),
                extra.map(v => chalk.dim(`[${v}]`)).join(' ')
            ];
        }));
    }
}

class Command {
    constructor(name, opt, handler) {
        invariant(typeof name === 'string', 'Expect command name to be a string');
        let names = [name];
        if (typeof opt === 'string') {
            opt = { desc: opt };
        }
        else if (typeof opt === 'object') {
            names = names.concat(opt.alias || []);
        }
        invariant(typeof opt.desc === 'string', 'Expect command description to be a string');
        const command = Object.assign({}, opt, { name, names: orderNames(names) });
        this.command = command;
        this.options = new Options();
        this.handler = handler;
    }
    option(name, opt) {
        this.options.add(name, opt);
        return this;
    }
}

var stripIndent = str => {
	const match = str.match(/^[ \t]*(?=\S)/gm);

	if (!match) {
		return str;
	}

	// TODO: use spread operator when targeting Node.js 6
	const indent = Math.min.apply(Math, match.map(x => x.length)); // eslint-disable-line
	const re = new RegExp(`^[ \\t]{${indent}}`, 'gm');

	return indent > 0 ? str.replace(re, '') : str;
};

var indentString = (str, count, opts) => {
	// Support older versions: use the third parameter as options.indent
	// TODO: Remove the workaround in the next major version
	const options = typeof opts === 'object' ? Object.assign({indent: ' '}, opts) : {indent: opts || ' '};
	count = count === undefined ? 1 : count;

	if (typeof str !== 'string') {
		throw new TypeError(`Expected \`input\` to be a \`string\`, got \`${typeof str}\``);
	}

	if (typeof count !== 'number') {
		throw new TypeError(`Expected \`count\` to be a \`number\`, got \`${typeof count}\``);
	}

	if (typeof options.indent !== 'string') {
		throw new TypeError(`Expected \`options.indent\` to be a \`string\`, got \`${typeof options.indent}\``);
	}

	if (count === 0) {
		return str;
	}

	const regex = options.includeEmptyLines ? /^/mg : /^(?!\s*$)/mg;
	return str.replace(regex, options.indent.repeat(count));
}
;

var redent = (str, count, indent) => indentString(stripIndent(str), count || 0, indent);

class Help {
    constructor(root, command, opts) {
        this.root = root;
        this.command = command;
        this.opts = opts;
    }
    getHelp() {
        let help = '\n';
        help += chalk.cyan(this.root.bin);
        if (this.root.pkg.version) {
            if (!this.opts.displayCommands && this.command) {
                help += ` ${this.command.command.name}`;
            }
            help += ` ${this.root.pkg.version}`;
        }
        help += '\n\n';
        if (!this.opts.displayCommands && this.command) {
            help += `${chalk.dim.italic(this.command.command.desc)}\n\n`;
        }
        else if (this.root.pkg.description) {
            help += `${chalk.dim.italic(this.root.pkg.description)}\n\n`;
        }
        const commandText = this.command
            ? chalk.magenta(this.opts.displayCommands
                ? '<command> '
                : `${this.command.command.name} `)
            : '';
        help += `${chalk.bold('USAGE')}\n\n`;
        help += redent(`${chalk.dim.italic(this.root.bin)} ${commandText}${chalk.yellow('[options]')}`, 2);
        help += '\n\n';
        if (this.opts.displayCommands && !this.root.isCommandsEmpty()) {
            help += formatSection({
                title: 'COMMANDS',
                body: this.root.commandsToString()
            });
        }
        if (this.command && !this.command.options.isEmpty()) {
            help += formatSection({
                title: 'COMMAND OPTIONS',
                body: this.command.options.toString()
            });
        }
        if (!this.root.options.isEmpty()) {
            help += formatSection({
                title: 'GLOBAL OPTIONS',
                body: this.root.options.toString()
            });
        }
        for (const h of this.root.extraHelps) {
            help += formatSection(h);
        }
        return redent(help, 2);
    }
    output() {
        process.stdout.write(this.getHelp());
        return this;
    }
}
function formatSection(sec) {
    if (typeof sec === 'string') {
        return sec;
    }
    const { title, body } = sec;
    return `${title ? `${chalk.bold(title)}\n\n` : ''}${redent(body.trim(), 2)}\n\n`;
}

var examplesPlugin = () => (cli) => {
    cli.on('parsed', command => {
        if (command && command.command.examples) {
            cli.extraHelp({
                title: 'EXAMPLES',
                body: command.command.examples.join('\n')
            });
        }
    });
};

var optionChoicesPlugin = () => (cli) => {
    cli.on('parsed', (command, _, flags) => {
        if (!command)
            return;
        const failedOptions = command.options.options.filter((option) => {
            const shouldCheck = option.choices && typeof flags[option.name] !== 'undefined';
            const isOneOfChices = option.choices &&
                option.choices.filter((choice) => {
                    return choice === flags[option.name];
                }).length > 0;
            return Boolean(shouldCheck && !isOneOfChices);
        });
        if (failedOptions.length > 0) {
            for (const option of failedOptions) {
                console.log(`The value of flag "${option.name}" should be one of: ${option.choices.map((choice) => `"${choice}"`)}`);
            }
            process.exit(1); // eslint-disable-line unicorn/no-process-exit
        }
    });
};

var requiredOptionPlugin = () => (cli) => {
    cli.on('parsed', (command, _, flags) => {
        if (!command)
            return;
        const missingRequiredOptions = command.options.options.filter(option => {
            const isRequired = option.required;
            const isMissing = typeof flags[option.name] === 'undefined';
            return isMissing && isRequired;
        });
        if (missingRequiredOptions.length > 0) {
            console.log(`Missing options: ${missingRequiredOptions
                .map(option => option.name)
                .join(', ')}`);
            process.exit(1); // eslint-disable-line unicorn/no-process-exit
        }
    });
};

// Prevent caching of this module so module.parent is always accurate
delete require.cache[__filename];
let parentDir;
if (module.parent) {
    parentDir = path.dirname(module.parent.filename);
}
else {
    parentDir = process.cwd();
}
class Cac extends EventEmitter {
    constructor({ bin, pkg, defaultOpts } = {}) {
        super();
        defaultOpts = defaultOpts || true;
        this.bin = bin || path.basename(process.argv[1]);
        this.commands = [];
        this.options = new Options();
        this.extraHelps = [];
        if (typeof defaultOpts === 'boolean') {
            this.helpOpt = defaultOpts;
            this.versionOpt = defaultOpts;
        }
        else if (typeof defaultOpts === 'object') {
            this.helpOpt = defaultOpts.help !== false;
            this.versionOpt = defaultOpts.version !== false;
        }
        this.pkg = Object.assign({}, pkg || readPkg.sync({ cwd: parentDir, normalize: false }).pkg);
        if (this.versionOpt) {
            this.option('version', {
                desc: 'Display version',
                alias: 'v',
                type: 'boolean'
            });
        }
        if (this.helpOpt) {
            this.option('help', {
                desc: `Display help (You're already here)`,
                alias: 'h',
                type: 'boolean'
            });
        }
        this.use(examplesPlugin());
        this.use(optionChoicesPlugin());
        this.use(requiredOptionPlugin());
    }
    /**
     * Use a plugin or an array of plugins
     */
    use(plugin) {
        if (Array.isArray(plugin)) {
            plugin.forEach(p => this.use(p));
        }
        else if (typeof plugin === 'function') {
            plugin(this);
        }
        else {
            throw new TypeError('plugin has to be a function or an array of it.');
        }
        return this;
    }
    /**
     * Add a global option
     */
    option(name, opt) {
        this.options.add(name, opt);
        return this;
    }
    /**
     * Add a sub command
     */
    command(name, opt, handler) {
        const command = new Command(name, opt, handler);
        this.commands.push(command);
        return command;
    }
    /**
     * Commands to string
     *
     * Used to display help
     */
    commandsToString() {
        return textTable$1(this.commands.map(({ command }) => {
            return [
                command.names.map(v => chalk.magenta(v)).join(', '),
                chalk.dim(command.desc)
            ];
        }));
    }
    /**
     * Check if there's any command
     */
    isCommandsEmpty() {
        return this.commands.length === 0;
    }
    /**
     * Find command by command name, alias or addtionalMatch
     */
    findCommand(name) {
        name = name || '*';
        for (const command of this.commands) {
            const { names, match } = command.command;
            if (names.includes(name)) {
                return { command, sliceFirstArg: Boolean(name && name !== '*') };
            }
            if (match && match(name)) {
                return { command, sliceFirstArg: false };
            }
        }
        return {
            command: null,
            sliceFirstArg: false
        };
    }
    get argv() {
        return this.parse(null, { run: false });
    }
    showHelp() {
        if (!this.started) {
            throw new Error('[cac] You have to call .parse() before running .showHelp()');
        }
        // Do not display `<command>` in help if it's a sub command
        const displayCommands = !(this.matchedCommand && isExplictCommand(this.firstArg));
        const help = new Help(this, this.matchedCommand, {
            displayCommands
        });
        help.output();
        return this;
    }
    /**
     * Show version in console
     */
    showVersion() {
        console.log(this.pkg.version);
    }
    /**
     * Add an extra help message
     */
    extraHelp(help) {
        this.extraHelps.push(help);
        return this;
    }
    /**
     * Parse CLI argument and run commands
     * @param argv Default to `process.argv.slice(2)`
     * @param opts
     */
    parse(argv, opts = {}) {
        const { run = true, showHelp } = opts;
        this.started = true;
        argv = argv || process.argv.slice(2);
        this.firstArg = argv[0] || '';
        // Ensure that first arg is not a flag
        this.firstArg = this.firstArg.startsWith('-') ? null : this.firstArg;
        const { command, sliceFirstArg } = this.findCommand(this.firstArg);
        this.matchedCommand = command;
        let { input, flags } = minimost(argv, {
            boolean: [
                ...this.options.getOptionNamesByType('boolean'),
                ...(command ? command.options.getOptionNamesByType('boolean') : [])
            ],
            string: [
                ...this.options.getOptionNamesByType('string'),
                ...(command ? command.options.getOptionNamesByType('string') : [])
            ],
            default: Object.assign({}, this.options.getDefaultsMapping(), (command ? command.options.getDefaultsMapping() : [])),
            alias: Object.assign({}, this.options.getAliasMap(), (command ? command.options.getAliasMap() : {}))
        });
        input = sliceFirstArg ? input.slice(1) : input;
        this.emit('parsed', command, input, flags);
        if (!run) {
            return { input, flags };
        }
        const shouldShowHelp = showHelp || (() => Boolean(this.helpOpt && flags.help));
        if (shouldShowHelp(command, input, flags)) {
            this.showHelp();
        }
        else if (this.versionOpt && flags.version) {
            this.showVersion();
        }
        else if (command && command.handler) {
            try {
                let res = command.handler(input, flags);
                Promise.resolve(res).then(() => {
                    this.emit('executed', command, input, flags);
                }).catch(this.handleError);
            }
            catch (err) {
                this.handleError(err);
            }
        }
        return { input, flags };
    }
    handleError(err) {
        if (this.listenerCount('error') === 0) {
            console.error(err.stack);
            process.exitCode = process.exitCode || 1;
        }
        else {
            this.emit('error', err);
        }
    }
}

/// <reference path="../declarations.d.ts" />
function cac(opts) {
    return new Cac(opts);
}
(function (cac) {
    function parse(args, opts) {
        return minimost(args, opts);
    }
    cac.parse = parse;
})(cac || (cac = {}));
var cac$1 = cac;

module.exports = cac$1;
